#include "DIDERO_PC.h"
#include <math.h>
#include <stdlib.h>
#include <windows.h>
#include <tchar.h>
#include <limits>
#include <stdio.h>
#include <string.h>

/* Mobius Global variable declarations, not definitions, are contained in the header files for the Atomic models */
//#include "../Atomic/ClientSAN/ClientSANSAN.h"
//#include "../Atomic/TransactionSAN_1/TransactionSAN_1SAN.h"
//#include "../Atomic/TransactionSAN_2/TransactionSAN_2SAN.h"
//#include "../Atomic/Initializer/InitializerSAN.h"

using namespace std;

char path_to_DIDEROPC_Project[MAX_PATH]={0};
char path_to_Client_plugin[MAX_PATH]={0};
char path_to_Server_plugin[MAX_PATH]={0};
char path_to_Server_plugin_2[MAX_PATH]={0};
char path_to_Middleware_plugin[MAX_PATH]={0};
char path_to_Debug_Output_File[MAX_PATH]={0};
char path_to_Client_debug_Output[MAX_PATH]={0};
char path_to_Server_debug_Output[MAX_PATH]={0};
char path_to_Server_debug_Output_2[MAX_PATH]={0};
char path_to_Middleware_debug_Output[MAX_PATH]={0};
char path_to_File_With_Emp_Dist_Files[MAX_PATH]={0};
char Trace_Server_Path[MAX_PATH]={0};
////char* Trace_Server_Path = "C:\\MobiusProject\\DIDERO_PC\\exes\\inter_proc_comm_server.exe";
char DIDERO_PC_Trace_Paths[4*MAX_PATH]={0};

/* Method for splitting paths from filenames */
std::string SplitFilename(const std::string& str);

//char* path_to_Client_plugin = "C:/MobiusProject/DIDERO_PC/DLLs/libDIDERO_PC_Client.dll";
//char* path_to_Server_plugin = "C:/MobiusProject/DIDERO_PC/DLLs/libDIDERO_PC_Server.dll";
//char* path_to_Server_plugin_2 = "C:/MobiusProject/DIDERO_PC/DLLs/libDIDERO_PC_Server_2.dll";
//char* path_to_Debug_Output_File = "C:/MobiusProject/DIDERO_PC/Output/Debug_Output_File.txt";
//char* path_to_Client_debug_Output = "C:/MobiusProject/DIDERO_PC/Output/libDIDERO_PC_Client_Output.txt";
//char* path_to_Server_debug_Output = "C:/MobiusProject/DIDERO_PC/Output/libDIDERO_PC_Server_Output.txt";
//char* path_to_Server_debug_Output_2 = "C:/MobiusProject/DIDERO_PC/Output/libDIDERO_PC_Server_Output_2.txt";
//char* path_to_File_With_Emp_Dist_Files = "C:/MobiusProject/DIDERO_PC/Input/Empirical_Dist_Paths.txt";
//char* Trace_Server_Path = "C:/Users/kizito/NetBeans_Projects/Inter_Proc_Comm_Server/dist/Debug/MinGW-Windows/inter_proc_comm_server.exe";
//char* DIDERO_PC_Trace_Paths = "C:/MobiusProject/DIDERO_PC/Output/DIDERO_PC_Client_Trace.txt C:/MobiusProject/DIDERO_PC/Output/DIDERO_PC_Server_1_Trace.txt C:/MobiusProject/DIDERO_PC/Output/DIDERO_PC_Server_2_Trace.txt C:/MobiusProject/DIDERO_PC/Output/DIDERO_PC_Simulation_summary.txt";

vector<Client_Interface*> ClientObjects;        /* Array of pointers to Client Object interface*/
Server_Interface* Server_1;        /* Pointer to Server Object interface*/
Server_Interface* Server_2;        /* Pointer to Server Object interface*/
Middleware_Interface* MW_ptr;      /* Pointer to Middleware Object interface */     
vector<bool> Exec_Oper_on_both_Servers;         /* Array of booleans to indicate that an operation 
                                                 * has been carried out on both servers. 
                                                 */
Inter_Proc_Comm_Client* Trace_data_Client;      /* Pointer to object that implements Client behaviour in a socket
                                                 * based solution for sending Trace data to a listening server 
                                                 */

vector<string> Oper_results;             /* Array of strings to store the result of the faster server */
map<int, map< int, map< int , string> > > Trans_Oper_Replica_Results;       /* Stores the results of operations executed by the 
Client_ID  Operation_ID   Server_ID                                            replicas. These results are compared by the Middlewware 
                                                                             * for possible inconsistencies. 
                                                                             */

map<int,int> Unbounded_Trans_ID;        /* Data structure to indicate how many transactions have been executed so far for a given Client */

map<int, map<int, vector<string> > > Oper_Data_for_Replicas;   /*Array to store the data that defines each operation in a Transaction */
/* Client_ID Operation_ID   Data */

map< int, Place* > Unblocked_place_ptrs;        /* Pointers to places that unblock blocked Transitions */
map< int, Place* > Blocked_place_ptrs;        /* Pointers to places that indicate the blocked status of each Transitions */

HMODULE DLL_ptr_Client;    /* Pointer to general interface for the "Client" DLLs */
HMODULE DLL_ptr_Server_1;    /* Pointer to general interface for the "Server" DLLs */
HMODULE DLL_ptr_Server_2;    /* Pointer to general interface for the "Server" DLLS*/   
HMODULE DLL_ptr_Middleware;    /* Pointer to general interface for the "Middleware" DLL*/   

create_Client create_Client_class;
destroy_Client destroy_Client_class;

create_Server create_Server_class;
destroy_Server destroy_Server_class;

create_Server create_Server_class_2;
destroy_Server destroy_Server_class_2;

create_Middleware create_Middleware_class;
destroy_Middleware destroy_Middleware_class;

vector<const char*> dlsym_error_vec;        /*TODO: May not need this vector since I only initialize a single plugin at any given time.
                                              Yes, I suspect this is "overkill"!! :-) Check.*/

int Experiment_Runs_Count;

/* The following data structure holds pairs of counts which, together, indicate how many times data is sent to the Trace logger */
/* The first Pair for the "client view", the second pair for the "Server_1 view" and the third for "Server_2"
 * 
 */

vector<int> ActionFiringCounts;


void Initialize_Clients_and_Servers()
{       
        /* Make sure that it is only at the beginning of an experiment Client and Server objects are created */

        static int Once_per_Experiment = 1;
        static int Experiment_Runs_Count_local = 1;
        
        
	if(Once_per_Experiment == 1)
	{
            Once_per_Experiment = 0;

            /* Obtain the location of this process. That is, where the ultimate process that calls this method
             * is spawned from on the harddrive. 
             */

            char buffer[MAX_PATH];//always use MAX_PATH for filepaths
            GetModuleFileName(NULL,buffer,sizeof(buffer));
            //std::cerr << "Filepath:" << buffer << "\n";

            stringstream temp_ss;
            stringstream temp_processpath;

            /* Store path to process*/
            /* Make sure to search for last backslash and ignore all string data after this */
            temp_processpath.str(std::string());
            temp_processpath << buffer;
            std::string processpath = SplitFilename(temp_processpath.str()); 
            std::replace(processpath.begin(), processpath.end(),'\\','/');
            //std::string processpath = SplitFilename(temp_processpath.str());

            temp_ss.str(std::string());
            temp_ss << processpath.c_str() << "/../../DLLs/libDIDERO_PC_Client.dll";
            ////std::cerr << temp_ss.str() << endl;
            //sprintf(path_to_Client_plugin, "%s", temp_ss.str().c_str());
            //memset(path_to_Client_plugin,'\0', MAX_PATH);
            temp_ss >> path_to_Client_plugin;
            ////std::cerr << processpath << endl;
            ////std::cerr << temp_ss.str() << endl;
            ////std::cerr << path_to_Client_plugin << endl;

            temp_ss.str(std::string());
            temp_ss.clear();
            ////std::cerr << processpath << endl;
            ////std::cerr << temp_ss.str() << endl;        
            temp_ss << processpath.c_str() << "/../../DLLs/libDIDERO_PC_Server.dll";
            //memset(path_to_Server_plugin,'\0', MAX_PATH);
            temp_ss >> path_to_Server_plugin;
            ////std::cerr << path_to_Server_plugin << endl;        

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../DLLs/libDIDERO_PC_Server_NoWait.dll";
            //memset(path_to_Server_plugin_2,'\0', MAX_PATH);
            temp_ss >> path_to_Server_plugin_2;        

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../DLLs/libDIDERO_PC_Middleware.dll";
            //memset(path_to_Middleware_plugin,'\0', MAX_PATH);
            temp_ss >> path_to_Middleware_plugin;
            
            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../Output/Debug_Output_File.txt";
            //memset(path_to_Debug_Output_File,'\0', MAX_PATH);
            temp_ss >> path_to_Debug_Output_File;        

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../Output/libDIDERO_PC_Client_Output.txt";
            //memset(path_to_Client_debug_Output,'\0', MAX_PATH);
            temp_ss >> path_to_Client_debug_Output;            

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../Output/libDIDERO_PC_Server_Output.txt";
            //memset(path_to_Server_debug_Output,'\0', MAX_PATH);
            temp_ss >> path_to_Server_debug_Output;           

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../Output/libDIDERO_PC_Server_Output_2.txt";
            //memset(path_to_Server_debug_Output_2,'\0', MAX_PATH);
            temp_ss >> path_to_Server_debug_Output_2;           
            
            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../Output/libDIDERO_PC_Middleware_Output.txt";
            //memset(path_to_Middleware_debug_Output,'\0', MAX_PATH);
            temp_ss >> path_to_Middleware_debug_Output;                       

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../Input/Empirical_Dist_Paths.txt";
            //memset(path_to_File_With_Emp_Dist_Files,'\0', MAX_PATH);
            temp_ss >> path_to_File_With_Emp_Dist_Files;
            ////std::cerr << path_to_File_With_Emp_Dist_Files << endl;

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../exes/inter_proc_comm_server.exe";
            //memset(Trace_Server_Path,'\0', MAX_PATH);
            temp_ss >> Trace_Server_Path;
            ////std::cerr << Trace_Server_Path << endl;

            temp_ss.str(std::string());
            temp_ss.clear();
            temp_ss << processpath.c_str() << "/../../Output/DIDERO_PC_Client_Trace.txt ";
            temp_ss << processpath.c_str() << "/../../Output/DIDERO_PC_Server_1_Trace.txt ";
            temp_ss << processpath.c_str() << "/../../Output/DIDERO_PC_Server_2_Trace.txt ";
            temp_ss << processpath.c_str() << "/../../Output/DIDERO_PC_Simulation_summary.txt";
            //memset(DIDERO_PC_Trace_Paths,'\0', 4*MAX_PATH);
            strncpy(DIDERO_PC_Trace_Paths, temp_ss.str().c_str(), sizeof(DIDERO_PC_Trace_Paths));               
            //std::cerr << temp_ss.str() <<" (c.f. the value of DIDERO_PC_Trace_Paths, which is " << DIDERO_PC_Trace_Paths << " )" << endl;            
                        
            
            /* Initialise global variables */
            create_Client_class = NULL;
            destroy_Client_class = NULL;
            
            create_Server_class = NULL;
            destroy_Server_class = NULL;
            
            create_Server_class_2 = NULL;
            destroy_Server_class_2 = NULL; 
            
            create_Middleware_class = NULL;
            destroy_Middleware_class = NULL;
            
            /* debug Output file stream */
            ofstream fileOut; 
            fileOut.open(path_to_Debug_Output_File,ios::out|ios::app);

            if(GV_debug_flag==1)
            {
                cerr << "An Output file stream has just been initialised" << endl;                
            }

            /******************* Create Client Objects to be used during simulation *****************************/
            /* Variable used for detecting errors/error handling */
            const char* dlsym_error = NULL;

            /* Initialize Global variables */
            ClientObjects.clear();        /* Array of pointers to Client Object interface*/
            Exec_Oper_on_both_Servers.clear();         /* Array of booleans to indicate that an operation */ 
            Oper_results.clear();             /* Array of strings to store the result of the faster server */
            Emp_Distributions.clear();          /* Array to hold the empirical distributions used during simulation */
            Sampled_SojournTimes.clear();       /* Data structure used to store the sampled sojourn times for each operation on each Server */
            Current_Num_of_Oper.clear();        /* Data structure to keep track of the current number of executing operations on each Server */
            PreviousEventTime.clear();          /* Data structure to keep track of the last sojourn-time sampling for each operation during simulation */
            Begin_Status.clear();                       /* Data structure to indicate when a "Begin" has occurred on the respective Server */
            Unbounded_Trans_ID.clear();        /* Data structure to indicate how many transactions have been executed so far for a given Client */
            Experiment_Runs_Count = Experiment_Runs_Count_local;
            ActionFiringCounts.clear();         /* The data structure holds pairs of counts which, together, indicate how many times data is sent to the Trace logger */
            
            /* Initialize ActionFiringCounts data structure for the 6 locations where trace data is sent to the Trace logger */
            for(int view_locations = 0; view_locations < 6; view_locations++)
            {
                ActionFiringCounts.push_back(0);
            }
            
            /* Create "Socket" object for sending Trace data to a listening Server */
            ////std::cerr << Trace_Server_Path << endl;
            temp_ss.str(std::string());
            temp_ss.clear();            
            temp_ss << Trace_Server_Path;
            Trace_data_Client = new Inter_Proc_Comm_Client(temp_ss.str());        
            //Trace_data_Client = new Inter_Proc_Comm_Client(std::string(Trace_Server_Path));

            //Send a 10000 lines of data
            //for(int i = 0; i<10000; i++)
            //{
            //    Trace_data_Client->send_test_data();
            //}
            
            //if(GV_debug_flag==15)
            //{
            //    if(dynamic_cast<SimulatorClass*>(&Simulator)==NULL)
            //    {
            //        std::cerr << "DIDERO_PC:: The \"this\" pointer does not refer the \"SimulatorClass\" object " << std::endl;
            //        
            //    }else
            //    {
            //        std::cerr << "DIDERO_PC:: The \"this\" pointer REFERS to the \"SimulatorClass\" object " << std::endl;
            //        std::cerr << "DIDERO_PC:: The values of MaxNumOfBatches and BatchRuns are " << Simulator.MaxNumOfBatches << " and " << Simulator.BatchRuns << std::endl;
            //    }
            //}
            
            /* Initialize the Empirical Distributions used during simulation */
            
            std::cerr << path_to_File_With_Emp_Dist_Files << endl;
            Init_Emp_Dists(string(path_to_File_With_Emp_Dist_Files));
            
            /* Initialise the server pointers to NULL */
            Server_1 = NULL;
            Server_2 = NULL;
            
            /* Initialize the Middleware pointer to NULL */
            MW_ptr = NULL;
            
            /*Obtain pointer to library interface*/
            //HMODULE testdll = LoadLibrary("libtestdll.dll") ;
            DLL_ptr_Client = NULL;

            if(GV_debug_flag==1)
            {
                fileOut << path_to_Client_plugin << " is the path to the Client plugin" <<endl;
            }

            DLL_ptr_Client = LoadLibrary(path_to_Client_plugin);

            if(DLL_ptr_Client!=NULL)
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL was successfully obtained " << endl;
                }
            }else
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL was NOT successfully obtained " << endl;
                }
            }            
            
            /* Obtain a pointer to a factory method in the DLL, using the method symbol which is its name,
             * whose interface is accessible via the plugin library.
             * So the "cast" below is defined in the interface's header file while the method "name" is 
             * in the related cpp file.
             */
            
            create_Client_class = (create_Client)GetProcAddress(DLL_ptr_Client, "create");
            
            if(create_Client_class!=NULL)
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL class method was successfully obtained " << endl;
                }
            }else
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL class method was NOT successfully obtained " << endl;
                }
            }
            
            /* Loop through the appropriate number of times, each time creating a suitable Client object.
            * Distribute the client objects across the warehouses as evenly as possible.
            *
            */

            for(int i = 0; i < GV_Number_of_Clients_and_Trans; i++)
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "About to create the " << i+1 <<"th Client." << endl;
                    fileOut << "W_ID: " << (i%GV_Number_of_Warehouses) + 1 <<", Number of Warehouses: " << GV_Number_of_Warehouses << ", Number of Districts per Warehouses: " << GV_Num_Dist_per_Warehouses << endl;
                }
                ClientObjects.push_back(NULL);
                ClientObjects[i] = create_Client_class(i+1 /*ID*/, (i%GV_Number_of_Warehouses) + 1 /*W_ID*/, GV_Number_of_Warehouses /*Number of Warehouses*/, GV_Num_Dist_per_Warehouses /* Number of Distracts per Warehouse */, path_to_Client_debug_Output, /* Debug output flag */GV_debug_flag, Prob_of_NewOrder, Prob_of_Payment, Prob_of_Orderstatus, Prob_of_Delivery, Prob_of_Stocklevel, Choose_Cust_using_ID, Prob_of_Local_Orderlines, Prob_of_Cust_ID, Prob_of_Bad_Cust);
                Exec_Oper_on_both_Servers.push_back(false);
                Oper_results.push_back(string(""));             /* Array of strings to store the result of the faster server */
                /* Even though the following data-structures are ultimately used to inform "Server-side" data/logic, the Servers do not know how many Clients they serve until the non-trivial
                 * part of the simulation begins, in which case they will create transactions according to Client requests for such. Consequently,
                 * it is in this sense that it is okay to have these initialized here */
                (Server_oper_index[i+1]).clear();               /* Array to store the progress of each Client Transaction on each Server replica */        
                (Trans_Oper_Replica_Results[i+1]).clear();      /* Array to store the results of each Servers executed operation in Client transactions */
                (Oper_Data_for_Replicas[i+1]).clear();          /* Array to store buffered operations from each Client */
                (Sampled_SojournTimes[1])[i+1] = 0.0;            /* Array to store the sampled values of sojourn-time for each operation on each Server */
                (Sampled_SojournTimes[2])[i+1] = 0.0;            /* Array to store the sampled values of sojourn-time for each operation on each Server */
                 PreviousEventTime[1][i+1] = 0.0;          /* Data structure to keep track of the last sojourn-time sampling for each operation during simulation */
                 PreviousEventTime[2][i+1] = 0.0;          /* Data structure to keep track of the last sojourn-time sampling for each operation during simulation */
                 (Begin_Status[1])[i+1] = false;                       /* Data structure to indicate when a "Begin" has occurred on the respective Server */
                 (Begin_Status[2])[i+1] = false;                       /* Data structure to indicate when a "Begin" has occurred on the respective Server */
                 Unbounded_Trans_ID[i+1] = 0;        /* Data structure to indicate how many transactions have been executed so far for a given Client */
            }
            
                 /* Initialize the data structure keeping track of how many operations are being performed currently on each Server */
                 Current_Num_of_Oper[1] = 0;
                 Current_Num_of_Oper[2] = 0;
            
            /********************************** Finished creating Client objects *******************************************/

            /********************************** Create Server and Middleware Objects *****************************************************/
            DLL_ptr_Server_1 = LoadLibrary(path_to_Server_plugin);
 
            if(DLL_ptr_Server_1!=NULL)
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL, DLL_ptr_Server_1, was successfully obtained " << endl;
                }
            }else
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL, DLL_ptr_Server_1, was NOT successfully obtained " << endl;                
                }
            }
            
            DLL_ptr_Server_2 = LoadLibrary(path_to_Server_plugin_2);
 
            if(DLL_ptr_Server_2!=NULL)
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL, DLL_ptr_Server_2, was successfully obtained " << endl;
                }
            }else
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL, DLL_ptr_Server_2, was NOT successfully obtained " << endl;                
                }
            }
            
            DLL_ptr_Middleware = LoadLibrary(path_to_Middleware_plugin);
 
            if(DLL_ptr_Middleware!=NULL)
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL, DLL_ptr_Middleware, was successfully obtained " << endl;
                }
            }else
            {
                if(GV_debug_flag==1)
                {
                    fileOut << "Pointer to DLL, DLL_ptr_Middleware, was NOT successfully obtained " << endl;                
                }
            }            
            
           /* Obtain a pointer to a factory method in the DLL, using the method symbol which is its name,
            * whose interface is accessible via the plugin library.
            * So the "cast" below is defined in the interface's header file while the method "name" is 
            * in the related cpp file.
            */
            
            if((create_Server_class)||(create_Server_class_2)||(create_Middleware_class))
            {
                std::cerr << "DIDERO_PC.cpp:: This code block should never be entered into" << std::endl;
                
            }else
            {
                create_Server_class = (create_Server) GetProcAddress(DLL_ptr_Server_1, "create");                
                create_Server_class_2 = (create_Server) GetProcAddress(DLL_ptr_Server_2, "create");                
                create_Middleware_class = (create_Middleware) GetProcAddress(DLL_ptr_Middleware, "create");
                
                if(create_Server_class!=NULL)
                {
                    if(GV_debug_flag==1)
                    {
                        fileOut << "DIDERO_PC.cpp:: Pointer to create method of DLL class method, create_Server_class, was successfully obtained" << endl;                        
                    }
                }else
                {
                    if(GV_debug_flag==1)
                    {
                        fileOut << "DIDERO_PC.cpp:: Pointer to create method of DLL class method, create_Server_class, was NOT successfully obtained " << endl;                
                    }
                }
                
                if(create_Server_class_2!=NULL)
                {
                    if(GV_debug_flag==1)
                    {
                        fileOut << "DIDERO_PC.cpp:: Pointer to create method of DLL class method, create_Server_class_2, was successfully obtained" << endl;                        
                    }
                }else
                {
                    if(GV_debug_flag==1)
                    {
                        fileOut << "DIDERO_PC.cpp:: Pointer to create method of DLL class method, create_Server_class_2, was NOT successfully obtained " << endl;                
                    }
                }
                
                if(create_Middleware_class!=NULL)
                {
                    if(GV_debug_flag==1)
                    {
                        fileOut << "DIDERO_PC.cpp:: Pointer to create method of DLL class method, create_Middleware_class, was successfully obtained" << endl;                        
                    }
                }else
                {
                    if(GV_debug_flag==1)
                    {
                        fileOut << "DIDERO_PC.cpp:: Pointer to create method of DLL class method, create_Middleware_class, was NOT successfully obtained " << endl;                
                    }
                }
            }

	    /* Loop through the appropriate number of times, each time creating a suitable Client object.
             * Distribute the client objects across the warehouses as evenly as possible.
             *
	     */
		 if(GV_debug_flag==1)
                 {
                     fileOut << "DIDERO_PC.cpp:: About to create the server objects." << endl;                    
                 }
		 Server_1 = (create_Server_class)(path_to_Server_debug_Output,GV_debug_flag);
		 Server_2 = (create_Server_class_2)(path_to_Server_debug_Output,GV_debug_flag);                 
                 int num_of_servers = static_cast<int>(GV_Server_1_enabled) + static_cast<int>(GV_Server_2_enabled);
                 MW_ptr = (create_Middleware_class)(GV_Number_of_Clients_and_Trans, 2, true, std::string(path_to_Middleware_debug_Output));
                 
                if((Server_1!=NULL)&&(Server_2!=NULL)&&(MW_ptr!=NULL))
                {
                    if(GV_debug_flag==1)
                    {
                        fileOut << "DIDERO_PC.cpp:: Success in Loading Server(s) and Middleware DLLs !!!" << endl;
                    }                    
                }
		
		fileOut.close();
				
                /***************************** End of Simulation Clients and Servers initialization ***************************
                 ************************************************************************************************/
	}else
	{
            /* Reset Global variables */
            Exec_Oper_on_both_Servers.clear();
            Oper_results.clear();             /* Array of strings to store the result of the faster server */
            Sampled_SojournTimes.clear();       /* Data structure used to store the sampled sojourn times for each operation on each Server */
            Current_Num_of_Oper.clear();
            PreviousEventTime.clear();          /* Data structure to keep track of the last sojourn-time sampling for each operation during simulation */
            Begin_Status.clear();                       /* Data structure to indicate when a "Begin" has occurred on the respective Server */
            Unbounded_Trans_ID.clear();        /* Data structure to indicate how many transactions have been executed so far for a given Client */
            
            /* If the number of experiment runs has been exceeded then free the socket connection for the Trace */
            Experiment_Runs_Count_local = Experiment_Runs_Count_local + 1;
            Experiment_Runs_Count = Experiment_Runs_Count_local;
            
            if((GV_Experiment_Runs<Experiment_Runs_Count)&&((Experiment_Runs_Count-GV_Experiment_Runs)==1))
            {
                std::cerr << "DIDERO_PC:: About to call destructor for Socket \"Client object\"!!!" << std::endl;
                delete Trace_data_Client;
            }

            
            /* Free DLLs */
            if(DLL_ptr_Server_1)
            {
                /* Unload the library from before */
            
                if(Server_1)
                {
                    if(destroy_Server_class)
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: There is no need to create a dynamic link to the server class, destroy_Server_class; it was created BEFORE" << endl;                            
                        }

                    }else
                    {
                        destroy_Server_class = (destroy_Server) GetProcAddress(DLL_ptr_Server_1, "destroy");                                    
                    }

                     if(GV_debug_flag==20)
                     {
                         cerr << "DIDERO_PC.cpp:: Server Class object is about to be destroyed " << endl;
                     }
                    
                    (destroy_Server_class)(Server_1);
                    Server_1 = NULL;
                }
                
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Server Library is about to be Freed!!! " << endl;
                }
                
                
                FreeLibrary(DLL_ptr_Server_1);
            }
            
            if(DLL_ptr_Server_2)
            {
                /* Unload the library from before */
            
                if(Server_2)
                {
                    if(destroy_Server_class_2)
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: There is no need to create a dynamic link to the server class, destroy_Server_class_2; it was created BEFORE" << endl;                            
                        }

                    }else
                    {
                        destroy_Server_class_2 = (destroy_Server) GetProcAddress(DLL_ptr_Server_2, "destroy");                                    
                    }

                     if(GV_debug_flag==20)
                     {
                         cerr << "DIDERO_PC.cpp:: Server Class object is about to be destroyed " << endl;
                     }
                    
                    (destroy_Server_class_2)(Server_2);
                    Server_2 = NULL;
                }
                
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Server Library is about to be Freed!!! " << endl;
                }
                
                FreeLibrary(DLL_ptr_Server_2);
            }
            
            if(DLL_ptr_Middleware)
            {
                /* Unload the library from before */
            
                if(MW_ptr)
                {
                    if(destroy_Middleware_class)
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: There is no need to create a dynamic link to the Middleware class, destroy_Middleware_class; it was created BEFORE" << endl;                            
                        }

                    }else
                    {
                        destroy_Middleware_class = (destroy_Middleware) GetProcAddress(DLL_ptr_Middleware, "destroy");                                    
                    }

                     if(GV_debug_flag==20)
                     {
                         cerr << "DIDERO_PC.cpp:: Middleware Class object is about to be destroyed " << endl;
                     }
                    
                    (destroy_Middleware_class)(MW_ptr);
                    MW_ptr = NULL;
                }
                
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Middleware Library is about to be Freed!!! " << endl;
                }
                
                FreeLibrary(DLL_ptr_Middleware);
            }
            
            DLL_ptr_Server_1 = NULL;
            create_Server_class = NULL;
            
            DLL_ptr_Server_2 = NULL;
            create_Server_class_2 = NULL;            
            
            DLL_ptr_Middleware = NULL;
            create_Middleware_class = NULL;                        
                       
            DLL_ptr_Server_1 = LoadLibrary(path_to_Server_plugin);
            DLL_ptr_Server_2 = LoadLibrary(path_to_Server_plugin_2);
            DLL_ptr_Middleware = LoadLibrary(path_to_Middleware_plugin);
            
            if(DLL_ptr_Server_1!=NULL)
            {
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Pointer to DLL class method, DLL_ptr_Server_1, was successfully obtained AGAIN " << endl;                    
                }

            }else
            {
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Pointer to DLL class method, DLL_ptr_Server_1, was NOT successfully obtained " << endl;                    
                }
            }
            
            if(DLL_ptr_Server_2!=NULL)
            {
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Pointer to DLL class method, DLL_ptr_Server_2, was successfully obtained AGAIN " << endl;                    
                }

            }else
            {
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Pointer to DLL class method, DLL_ptr_Server_2, was NOT successfully obtained " << endl;                    
                }
            }

            if(DLL_ptr_Middleware!=NULL)
            {
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Pointer to DLL class method, DLL_ptr_Middleware, was successfully obtained AGAIN " << endl;                    
                }

            }else
            {
                if(GV_debug_flag==20)
                {
                    cerr << "DIDERO_PC.cpp:: Pointer to DLL class method, DLL_ptr_Middleware, was NOT successfully obtained " << endl;                    
                }
            }            
            
            /* debug Output file stream */
            //ofstream fileOut; 
            //fileOut.open(path_to_Debug_Output_File,ios::out|ios::app);                        
               
                /* Obtain a pointer to a factory method in the DLL, using the method symbol which is its name,
                * whose interface is accessible via the plugin library.
                * So the "cast" below is defined in the interface's header file while the method "name" is 
                * in the related cpp file.
                */
                
                if(create_Server_class)
                {
                    if(GV_debug_flag==20)
                    {
                        cerr << "DIDERO_PC.cpp:: No need to create a dynamic link to a Server class, create_Server_class; it was already created BEFORE" << endl;                        
                    }

                }else
                {
                    create_Server_class = (create_Server) GetProcAddress(DLL_ptr_Server_1, "create");

                    if(create_Server_class!=NULL)
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: Pointer to method of DLL class method, create_Server_class, was successfully obtained AGAIN " << endl;                            
                        }

                    }else
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: Pointer to method of DLL class method, create_Server_class, was NOT successfully obtained " << endl;                                            
                        }

                    }                                        
                }

                if(create_Server_class_2)
                {
                    if(GV_debug_flag==20)
                    {
                        cerr << "DIDERO_PC.cpp:: No need to create a dynamic link to a Server class, create_Server_class_2; it was already created BEFORE" << endl;                        
                    }

                }else
                {
                    create_Server_class_2 = (create_Server) GetProcAddress(DLL_ptr_Server_2, "create");

                    if(create_Server_class_2!=NULL)
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: Pointer to method of DLL class method, create_Server_class_2, was successfully obtained AGAIN " << endl;                            
                        }

                    }else
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: Pointer to method of DLL class, create_Server_class_2, was NOT successfully obtained " << endl;                                            
                        }

                    }                                        
                }            
            
                if(create_Middleware_class)
                {
                    if(GV_debug_flag==20)
                    {
                        cerr << "DIDERO_PC.cpp:: No need to create a dynamic link to a Middleware class, create_Middleware_class; it was already created BEFORE" << endl;                        
                    }

                }else
                {
                    create_Middleware_class = (create_Middleware) GetProcAddress(DLL_ptr_Middleware, "create");

                    if(create_Middleware_class!=NULL)
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: Pointer to method of DLL class method, create_Middleware_class, was successfully obtained AGAIN " << endl;                            
                        }

                    }else
                    {
                        if(GV_debug_flag==20)
                        {
                            cerr << "DIDERO_PC.cpp:: Pointer to method of DLL class method, create_Middleware_class, was NOT successfully obtained " << endl;                                            
                        }
                    }                                        
                }
            
                /* Loop through the appropriate number of times, each time creating a suitable Client object.
                 * Distribute the client objects across the warehouses as evenly as possible.
                 *
                 */

                 if(GV_debug_flag==20)
                 {
                     cerr << "About to create the server objects." << endl;                                         
                 }

                 Server_1 = (create_Server_class)(path_to_Server_debug_Output,GV_debug_flag);
                 Server_2 = (create_Server_class_2)(path_to_Server_debug_Output,GV_debug_flag);                 
                 int num_of_servers = static_cast<int>(GV_Server_1_enabled) + static_cast<int>(GV_Server_2_enabled);
                 MW_ptr = (create_Middleware_class)(GV_Number_of_Clients_and_Trans, 2, true, std::string(path_to_Middleware_debug_Output));
                 
                 
                if((Server_1!=NULL)&&(Server_2!=NULL)&&(MW_ptr!=NULL))
                {
                    if(GV_debug_flag==20)
                    {
                        cerr << "DIDERO_PC.cpp:: Success in RE-Loading Server(s) and Middleware DLLs !!!" << endl;
                    }                    
                }                 
                 
                 
                 /* Create Transaction/Connection objects for each Server and map them to the already created Client objects */
                 //for(int i = 0; i < GV_Number_of_Clients_and_Trans; i++)
                 //{
                 //    fileOut << "About to create the "<< i+1 <<"th connection object for each server." << endl;                                        
                 //    Server_1->create_transaction(i+1);
                 //}
            
            if(DLL_ptr_Client)
            {
                for(int j = 0; j < GV_Number_of_Clients_and_Trans; j++)
                {
                    /* No need to grab another pointer to the DLL */
                    if(destroy_Client_class)
                    {
                        if(GV_debug_flag==1)
                        {
                            cerr << "No need to get pointer to destroy method for Client objects; it was already obtained BEFORE" << endl;                            
                        }

                    }else
                    {
                        destroy_Client_class = (destroy_Client) GetProcAddress(DLL_ptr_Client, "destroy");
                    }
                    
                    if(ClientObjects[j])
                    {
                        (destroy_Client_class)(ClientObjects[j]);
                        ClientObjects[j] = NULL;
                    }                        
                    
                }
                
                if(GV_debug_flag==1)
                {
                    cerr << " Client library is about to be freed!!!! " << endl;
                }
                
                FreeLibrary(DLL_ptr_Client);
            }
                 DLL_ptr_Client = NULL;
                 create_Client_class = NULL;
                
                for(int j = 0; j < GV_Number_of_Clients_and_Trans; j++)
                {
                    if(DLL_ptr_Client)
                    {
                        /* No need to obtain pointer */
                    }else
                    {
                        DLL_ptr_Client = LoadLibrary(path_to_Client_plugin);

                        if(DLL_ptr_Client!=NULL)
                        {
                            if(GV_debug_flag==20)
                            {
                                cerr << "Pointer to DLL was successfully obtained AGAIN" << endl;                                
                            }

                        }else
                        {
                            if(GV_debug_flag==20)
                            {
                                cerr << "Pointer to DLL was NOT successfully obtained " << endl;                                
                            }

                        }                        
                    }
                    
                    if(GV_debug_flag==20)
                    {
                        cerr << path_to_Client_plugin << " is the path to the Client plugin" <<endl;                        
                    }


                    /* Obtain a pointer to a factory method in the DLL, using the method symbol which is its name,
                     * whose interface is accessible via the plugin library.
                     * So the "cast" below is defined in the interface's header file while the method "name" is 
                     * in the related cpp file.
                     */

                    if(create_Client_class)
                    {
                        /* No need to create create_Client_class pointer */
                    }else
                    {
                        create_Client_class = (create_Client)GetProcAddress(DLL_ptr_Client, "create");
                        
                        if(create_Client_class!=NULL)
                        {
                            if(GV_debug_flag==20)
                            {
                                cerr << "Pointer to DLL class method was successfully obtained " << endl;                                
                            }

                        }else
                        {
                            if(GV_debug_flag==20)
                            {
                                cerr << "Pointer to DLL class method was NOT successfully obtained " << endl;                                
                            }

                        }                        
                    }

                     /* Loop through the appropriate number of times, each time creating a suitable Client object.
                      * Distribute the client objects across the warehouses as evenly as possible.
                      *
                      */

                    if(GV_debug_flag==20)
                    {
                        cerr << "About to create the " << j+1 <<"th Client." << endl;
                        cerr << "W_ID: " << (j%GV_Number_of_Warehouses) + 1 <<", Number of Warehouses: " << GV_Number_of_Warehouses << ", Number of Districts per Warehouses: " << GV_Num_Dist_per_Warehouses << endl;                        
                    }

                     ClientObjects[j] = create_Client_class(j+1 /*ID*/, (j%GV_Number_of_Warehouses) + 1 /*W_ID*/, GV_Number_of_Warehouses /*Number of Warehouses*/, GV_Num_Dist_per_Warehouses /* Number of Distracts per Warehouse */, path_to_Client_debug_Output, /* Debug output flag */ GV_debug_flag, Prob_of_NewOrder, Prob_of_Payment, Prob_of_Orderstatus, Prob_of_Delivery, Prob_of_Stocklevel, Choose_Cust_using_ID, Prob_of_Local_Orderlines, Prob_of_Cust_ID, Prob_of_Bad_Cust);
                     Exec_Oper_on_both_Servers.push_back(false);
                     Oper_results.push_back(string(""));             /* Array of strings to store the result of the faster server */
                     /* Even though the following data-structures are ultimately used to inform "Server-side" data/logic, the Servers do not know how many Clients they serve until the non-trivial
                        * part of the simulation begins, in which case they will create transactions according to Client requests for such. Consequently,
                        * it is in this sense that it is okay to have these initialized here */
                        (Server_oper_index[j+1]).clear();               /* Array to store the progress of each Client Transaction on each Server replica */        
                        (Trans_Oper_Replica_Results[j+1]).clear();      /* Array to store the results of each Servers executed operation in Client transactions */
                        (Oper_Data_for_Replicas[j+1]).clear();          /* Array to store buffered operations from each Client */
                        (Sampled_SojournTimes[1])[j+1] = 0.0;            /* Array to store the sampled values of sojourn-time for each operation on each Server */
                        (Sampled_SojournTimes[2])[j+1] = 0.0;            /* Array to store the sampled values of sojourn-time for each operation on each Server */
                        PreviousEventTime[1][j+1] = 0.0;          /* Data structure to keep track of the last sojourn-time sampling for each operation during simulation */
                        PreviousEventTime[2][j+1] = 0.0;          /* Data structure to keep track of the last sojourn-time sampling for each operation during simulation */
                        (Begin_Status[1])[j+1] = false;                       /* Data structure to indicate when a "Begin" has occurred on the respective Server */
                        (Begin_Status[2])[j+1] = false;                       /* Data structure to indicate when a "Begin" has occurred on the respective Server */
                        Unbounded_Trans_ID[j+1] = 0;        /* Data structure to indicate how many transactions have been executed so far for a given Client */
                }
                 
                 /* Initialize the data structure keeping track of how many operations are being performed currently on each Server */
                 Current_Num_of_Oper[1] = 0;
                 Current_Num_of_Oper[2] = 0;
                 
                 //fileOut.close();    
	}
}

std::string SplitFilename(const std::string& str)
{
  //std::cerr << "Splitting: " << str << '\n';
  unsigned found = str.find_last_of("/\\");
  //std::cerr << " path: " << str.substr(0,found) << '\n';
  //std::cerr << " file: " << str.substr(found+1) << '\n';
  return str.substr(0,found);
}
