/* 
 * File:   Client_Interface.h
 * Author: kizito
 *
 * Created on 23 October 2013, 15:36
 */

#ifndef MIDDLEWARE_INTERFACE_H
#define	MIDDLEWARE_INTERFACE_H

#ifdef MY_COMPILE_DLL
#define MY_DECLSPEC __declspec(dllexport)
#else
#define MY_DECLSPEC __declspec(dllimport)
#endif


#include <iostream>
#include <string>

using namespace std;

class MY_DECLSPEC Middleware_Interface
{

        public:
        
        virtual ~Middleware_Interface(){};
        
        /* Store the current operation sent by the Client */
        virtual int MW_buffer_operation(int, std::string, std::string, int) = 0;
        /* Reset the Transaction progress buffer related to a given Client */
        virtual int MW_buffer_reset(int) = 0;
        /* Reset buffer for storing results of operations executed for the 
         * relevant client 
         */
        virtual int MW_results_reset(int) = 0;
        /* Get the current operation key for the current operation
         * from the relevant client for the relevant server
         */
        virtual std::string Server_get_operation_data(int,int) = 0;
        /* get the current operation from the relevant client for the relevant
         * server
         */
        virtual std::string Server_get_operation(int,int) = 0;
        /* Set the result of performing the current operation for the
         * relevant client
         */
        virtual int Server_set_operation_result(int,int,std::string) = 0;
        /* Get the result of performing the current operation for the
         * relevant client
         */
        virtual std::string MW_get_operation_result(int,int) = 0;
        /* Detect whether the current operation for the given client is an edge */
        virtual bool MW_detect_edge(int/* Client_ID*/) = 0;
        /* Signal faster Server */
        virtual int MW_signal_server(int) = 0;
        /* Detect Exception */
        virtual bool MW_detect_exception(int/*Client_ID*/,int/*Server_ID*/) = 0;
        /* Detect Successful transaction termination */
        virtual bool MW_detect_Transaction_complete(int/*Client_ID*/,int/*Server_ID*/) = 0;
        /* Get Transaction operation ID */
        virtual int Server_get_Trans_oper_ID(int/* Client_ID*/,int/*Server_ID*/)=0;
        /* Update Transaction progress for the given server */
        virtual int MW_update_Trans_progress(int/* Client_ID*/,int/*Server_ID*/)=0;
        /* Get Transaction Progress */
        virtual int MW_get_Trans_progress(int/* Client_ID*/,int/*Server_ID*/)=0;
};

    typedef Middleware_Interface* (*create_Middleware)(int /* Number of Clients */, int /* Number of Servers*/, bool /* debug flag*/, std::string /* debug file path */);               /*There should be a method, implemented in the same source file as the plugin
                                                                     class definition (the most specialised/"derived" class in the
                                                                     *plugin hierarchy), whose prototype matches
                                                                     this typedef and the purpose of which is to instantiate an
                                                                     object of the class--type that is "highest"/leaf/"most specialised"
                                                                     in the hierarchy*/
    typedef void (*destroy_Middleware)(Middleware_Interface*);
    

#endif	/* MIDDLEWARE_INTERFACE_H */
