#include <algorithm>
#include <map>
#include "DIDERO_PC.h"

/************ Header files for Boost Distributions *************/
/* I think the following header contains the typedef for minstd_rand */
#include <boost/random/linear_congruential.hpp>

#include "boost/random.hpp"

//#include <boost/random/exponential_distribution.hpp>
#include <boost/random/uniform_real.hpp>

/* I think this header file is for the "definition" of the "variate_generator" data-type  */
#include <boost/random/variate_generator.hpp>

/* KS comment (17-08-2010: )The following appears to be a typedef of a "data-type" for boost's random number seed generators */
typedef boost::minstd_rand base_generator_type;

/* Initialize random number generation functions */
static base_generator_type generator(42u);

/* Data-structure used to store estimated sojourn-times for each operation on each server */
map< int, map< int, double > > previous_sampled_sojtimes;

/* Inverse Transform Sampling */
boost::uniform_real<double> InverseTransformSample(0.0, 1.0);
boost::variate_generator<base_generator_type&, boost::uniform_real<double> > Inverse_Transform_Sample(generator, InverseTransformSample);          


using namespace std;

double sample(map<double, double>* dist, bool new_quantile, bool reset_previous_quantile, int Trans_ID, int Server_ID, double current_sim_time, bool reactivate)
{     
    double duration = 0.0;
    
    static map< int, map< int, double > > previous_quantiles;
    static map< int, map< int, double > > Transaction_quantiles;
    static bool init_previous_quantiles = true;
    
    /* Initialise the "previous_quantiles" map once */
    if(init_previous_quantiles)
    {
        previous_quantiles.clear();
        previous_sampled_sojtimes.clear();
        Transaction_quantiles.clear();
        init_previous_quantiles = false;
    }
    
    map<double, double>::iterator iter;
    map<double, double>::iterator iter_lower;
    
    double u_dist = 0.0;
        
    if(new_quantile)
    {
        if(!GV_Couple_Server_Trans_Durations)
        {
                u_dist = Inverse_Transform_Sample();
                (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                (Transaction_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
        }else
        {
            if((previous_quantiles[Trans_ID-1])[0] != 0.0)
            {
                u_dist = (previous_quantiles[Trans_ID-1])[0];
                (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                (Transaction_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;

            }else if((previous_quantiles[Trans_ID-1])[1] != 0.0)
            {
                u_dist = (previous_quantiles[Trans_ID-1])[1];  
                (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                (Transaction_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;

            }else
            {
                u_dist = Inverse_Transform_Sample();
                (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                (Transaction_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
            }            
        }

    }else
    {
        if(Server_ID==1)
        {
            /* NOWAIT-disabled Server */
            if((GV_IntraTransaction_Dependence)&&(GV_Server_1_Prob_of_IntraTransDependence > Inverse_Transform_Sample())&&(!GV_DefaultResidualTimes))
            {
                /* Intra-Transaction dependence is enabled */
                if(((0.99999 > (previous_quantiles[Trans_ID-1])[Server_ID-1])&&(0.979 < (previous_quantiles[Trans_ID-1])[Server_ID-1]))||((0.98 > (previous_quantiles[Trans_ID-1])[Server_ID-1])&&(0.0 < (previous_quantiles[Trans_ID-1])[Server_ID-1])))
                {
                    //std::numeric_limits<double>::epsilon()
                    u_dist = (previous_quantiles[Trans_ID-1])[Server_ID-1];            

                }else
                {
                    u_dist = Inverse_Transform_Sample();
                    if((!GV_Couple_Server_Trans_Durations)&&(GV_IntraTransDepBurstiness))
                    {
                        (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;                        
                    }
                }
                
            }else if(GV_DefaultResidualTimes)
            {
                if(reactivate)
                {
                    u_dist = (previous_quantiles[Trans_ID-1])[Server_ID-1];                                                 
                
                }else
                {
                    if(GV_Server_1_Prob_of_IntraTransDependence > Inverse_Transform_Sample())
                    {
                        (previous_quantiles[Trans_ID-1])[Server_ID-1] = (Transaction_quantiles[Trans_ID-1])[Server_ID-1];
                        u_dist = (previous_quantiles[Trans_ID-1])[Server_ID-1];
                    
                    }else
                    {
                        u_dist = Inverse_Transform_Sample();
                        (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                    }
                }
            
            }else
            {
                /* Independent Intra-Transaction sampling case */
                u_dist = Inverse_Transform_Sample();
                if((!GV_Couple_Server_Trans_Durations)&&(GV_IntraTransaction_Dependence)&&(GV_IntraTransDepBurstiness))
                {
                    (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                }
            }
            

        }else if(Server_ID==2)
        {
            /* NOWAIT Server */
            if((GV_IntraTransaction_Dependence)&&(GV_Server_2_Prob_of_IntraTransDependence > Inverse_Transform_Sample())&&(!GV_DefaultResidualTimes))
            {
                if((((0.9999999999999999999 > (previous_quantiles[Trans_ID-1])[Server_ID-1])&&(0.9999 < (previous_quantiles[Trans_ID-1])[Server_ID-1]))||((0.978579 > (previous_quantiles[Trans_ID-1])[Server_ID-1])&&(0.23 < (previous_quantiles[Trans_ID-1])[Server_ID-1])))&&(GV_IntraTransaction_Dependence))
                {
                    //std::numeric_limits<double>::epsilon()
                    u_dist = (previous_quantiles[Trans_ID-1])[Server_ID-1];            

                }else
                {
                    u_dist = Inverse_Transform_Sample();
                    if((!GV_Couple_Server_Trans_Durations)&&(GV_IntraTransDepBurstiness))
                    {
                        (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;                        
                    }
                }                            
            }else if(GV_DefaultResidualTimes)
            {
                if(reactivate)
                {
                    u_dist = (previous_quantiles[Trans_ID-1])[Server_ID-1];                                                 
                
                }else
                {
                    if(GV_Server_2_Prob_of_IntraTransDependence > Inverse_Transform_Sample())
                    {
                        (previous_quantiles[Trans_ID-1])[Server_ID-1] = (Transaction_quantiles[Trans_ID-1])[Server_ID-1];
                        u_dist = (previous_quantiles[Trans_ID-1])[Server_ID-1];
                    
                    }else
                    {
                        u_dist = Inverse_Transform_Sample();
                        (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                    }
                }
            }else
            {
                /* Independent Intra-Transaction sampling case */
                u_dist = Inverse_Transform_Sample();
                if((!GV_Couple_Server_Trans_Durations)&&(GV_IntraTransaction_Dependence)&&(GV_IntraTransDepBurstiness))
                {
                    (previous_quantiles[Trans_ID-1])[Server_ID-1] = u_dist;
                }
            }
        }
    }

    
    if(GV_debug_flag==5)
    {
        std::cerr << "Sample_Emp_Dist.cpp:: A uniformly distributed value of "<< u_dist << " has been sampled!!!!" << std::endl;
        std::cerr << "Sample_Emp_Dist.cpp:: ************** Printing Distribution ************* " << std::endl;
        for(map<double, double>::iterator iter_2 = dist->begin(); iter_2 != dist->end(); iter_2++)
        {
                std::cerr << "Sample_Emp_Dist.cpp:: Prob: " << (*iter_2).first << " Value: " << (*iter_2).second << std::endl;            
        }
        std::cerr << "Sample_Emp_Dist.cpp:: **************** End of Distribution ****************** " << std::endl;
    }
    
    iter = dist->lower_bound(u_dist);
    
    /* Proceed, depending on whether the inverse-transform algorithm requires linear
     * interpolation or not  
     */
    
    if(GV_Step_Inverse_Transform)
    {
        duration = (*iter).second;        
    
    }else
    {
        /* If the quantile exists explicitly, then conduct no interpolation.
         * Otherwise, linearly interpolate between the quantile returned and a quantile lower.
         * The lower quantile must exist since the "0" and "1" quantiles -- the "edges" of the ECDF --
         * must exist for a well defined ECDF
         */
        if(iter->first==u_dist)
        {
            duration = (*iter).second;        

        }else
        {
            /* Obtain quantile just below */
            iter_lower = iter;        
            iter_lower--;

            /* Linearly interpolate between the quantiles */
            duration = iter_lower->second + ((u_dist-iter_lower->first)/(iter->first-iter_lower->first))*(iter->second-iter_lower->second);        
        }        
    }
    
    /* If necessary, reset the last quantile stored */
    if(reset_previous_quantile)
    {
        (previous_quantiles[Trans_ID-1])[Server_ID-1] = 0.0;                    
    }
    
    /* Store the time of the last duration */
    if(GV_DefaultResidualTimes)
    {
        /* If this is a new sojourn-time then return the entire duration */
        if(!reactivate)
        {
            (previous_sampled_sojtimes[Trans_ID-1])[Server_ID-1] = duration + current_sim_time;            
        }else
        {
            /* Otherwise, return the residual */
            duration = (previous_sampled_sojtimes[Trans_ID-1])[Server_ID-1] - current_sim_time;
        }        
    }
    
    return duration;
}

double estimate_Percentile(map<double, double>* dist, double percentile)
{
    double sample_value = 0;
    double apprx_percentile = 0;
    
    for(int i = 0; i <= 10000000; i++)
    {
        sample_value = sample(dist, true, false, 1, 1, 1.0, true);
        if(sample_value>percentile)
        {
            /* Do not increase the apprx_percentile */
        }else
        {
           apprx_percentile = apprx_percentile + 1; 
        }
    }
    
    apprx_percentile = apprx_percentile/10000000;
    
    return apprx_percentile;
}
